/**
 *
 * Compiz Autumn plugin
 *
 * autumn.c
 * 
 * Autumn was adapted from the Snow plugin by Patrick Fisher. The copywrites and warranty 
 * from the Snow plugin remain in effect, as shown below.
 *
 * Copyright (c) 2006 Eckhart P. <beryl@cornergraf.net>
 * Copyright (c) 2006 Brian Jørgensen <qte@fundanemt.com>
 * Snow plugin maintained by Danny Baumann <maniac@opencompositing.org>
 * Autumn Plugin maintained by Patrick Fisher <autumnplugin@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 **/

/*
 * Many thanks to Atie H. <atie.at.matrix@gmail.com> for providing
 * a clean plugin template
 * Also thanks to the folks from #beryl-dev, especially Quinn_Storm
 * for helping me make this possible
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

#include <X11/Xatom.h>
#include <X11/extensions/Xrender.h>

#include <compiz.h>
#include "autumn_options.h"
#include <time.h>


#define GET_AUTUMN_DISPLAY(d)                            \
    ((AutumnDisplay *) (d)->privates[displayPrivateIndex].ptr)

#define AUTUMN_DISPLAY(d)                                \
    AutumnDisplay *sd = GET_AUTUMN_DISPLAY (d)

#define GET_AUTUMN_SCREEN(s, sd)                         \
    ((LeafScreen *) (s)->privates[(sd)->screenPrivateIndex].ptr)

#define AUTUMN_SCREEN(s)                                 \
    LeafScreen *ss = GET_AUTUMN_SCREEN (s, GET_AUTUMN_DISPLAY (s->display))

static int displayPrivateIndex = 0;

/* -------------------  STRUCTS ----------------------------- */
typedef struct _AutumnDisplay
{
    int screenPrivateIndex;

    Bool useTextures;

    int             autumnTexNFiles;
    CompOptionValue *autumnTexFiles;
} AutumnDisplay;

typedef struct _LeafTexture
{
    CompTexture tex;

    unsigned int width;
    unsigned int height;

    Bool   loaded;
    GLuint dList;
} LeafTexture;

typedef struct _Leaf
{
    float x, y, z;
    float xs, ys, zs;
    float ra; /* rotation angle */
    float rs; /* rotation speed */

    LeafTexture *tex;
} Leaf;

typedef struct _LeafScreen
{
    CompScreen *s;

    Bool active;

    CompTimeoutHandle timeoutHandle;

    PaintOutputProc paintOutput;
    DrawWindowProc  drawWindow;

    LeafTexture *autumnTex;
    int         autumnTexturesLoaded;

    GLuint displayList;
    Bool   displayListNeedsUpdate;

    Leaf *allLeaves;
} LeafScreen;

/* some forward declarations */
static void initiateLeaf (LeafScreen * ss, Leaf * sf);
static void autumnMove (CompDisplay *d, Leaf * sf);

static void
autumnThink (LeafScreen *ss,
	   Leaf  *sf)
{
    int boxing;

    boxing = autumnGetScreenBoxing (ss->s->display);

    if (sf->y >= ss->s->height + boxing ||
	sf->x <= -boxing ||
	sf->y >= ss->s->width + boxing ||
	sf->z <= -((float) autumnGetScreenDepth (ss->s->display) / 500.0) ||
	sf->z >= 1)
    {
	initiateLeaf (ss, sf);
    }
    autumnMove (ss->s->display, sf);
}

static void
autumnMove (CompDisplay *d,
	  Leaf   *sf)
{
    float tmp = 1.0f / (100.0f - autumnGetWindSpeed (d));
    int   autumnUpdateDelay = autumnGetAutumnUpdateDelay (d);

    sf->x += (sf->xs * (float) autumnUpdateDelay) * tmp;
    sf->y += (sf->ys * (float) autumnUpdateDelay) * tmp;
    sf->z += (sf->zs * (float) autumnUpdateDelay) * tmp;
    sf->ra += ((float) autumnUpdateDelay) / (10.0f - sf->rs);
}

static Bool
stepLeafPositions (void *closure)
{
    CompScreen *s = closure;
    int        i, numLeaves;
    Leaf  *autumnLeaf;
    Bool       onTop;

    AUTUMN_SCREEN (s);

    if (!ss->active)
	return TRUE;

    autumnLeaf = ss->allLeaves;
    numLeaves = autumnGetNumLeaves (s->display);
    onTop = autumnGetLeavesOverWindows (s->display);

    for (i = 0; i < numLeaves; i++)
	autumnThink(ss, autumnLeaf++);

    if (ss->active && !onTop)
    {
	CompWindow *w;

	for (w = s->windows; w; w = w->next)
	{
	    if (w->type & CompWindowTypeDesktopMask)
		addWindowDamage (w);
	}
    }
    else if (ss->active)
	damageScreen (s);

    return TRUE;
}

static Bool
autumnToggle (CompDisplay     *d,
	    CompAction      *action,
	    CompActionState state,
 	    CompOption      *option,
	    int             nOption)
{
    CompScreen *s;
    Window     xid;

    xid = getIntOptionNamed (option, nOption, "root", 0);
    s = findScreenAtDisplay (d, xid);

    if (s)
    {
	AUTUMN_SCREEN (s);
	ss->active = !ss->active;
	if (!ss->active)
	    damageScreen (s);
    }

    return TRUE;
}

/* --------------------  HELPER FUNCTIONS ------------------------ */

static inline int
getRand (int min,
	 int max)
{

    return (rand() % (max - min + 1) + min);
}

static inline float
mmRand (int   min,
	int   max,
	float divisor)
{
    return ((float) getRand(min, max)) / divisor;
};

/* --------------------------- RENDERING ------------------------- */
static void
setupDisplayList (LeafScreen *ss)
{
    float autumnSize = autumnGetLeafSize (ss->s->display);

    ss->displayList = glGenLists (1);

    glNewList (ss->displayList, GL_COMPILE);
    glBegin (GL_QUADS);

    glColor4f (1.0, 1.0, 1.0, 1.0);
    glVertex3f (0, 0, -0.0);
    glColor4f (1.0, 1.0, 1.0, 1.0);
    glVertex3f (0, autumnSize, -0.0);
    glColor4f (1.0, 1.0, 1.0, 1.0);
    glVertex3f (autumnSize, autumnSize, -0.0);
    glColor4f (1.0, 1.0, 1.0, 1.0);
    glVertex3f (autumnSize, 0, -0.0);

    glEnd ();
    glEndList ();
}

static void
beginRendering (LeafScreen *ss,
		CompScreen *s)
{

    if (autumnGetUseBlending (s->display))
	glEnable (GL_BLEND);

    glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

    if (ss->displayListNeedsUpdate)
    {
	setupDisplayList (ss);
	ss->displayListNeedsUpdate = FALSE;
    }

    glColor4f (1.0, 1.0, 1.0, 1.0);
    if (ss->autumnTexturesLoaded && autumnGetUseTextures (s->display))
    {
	int j;

	for (j = 0; j < ss->autumnTexturesLoaded; j++)
	{
	    Leaf *autumnLeaf = ss->allLeaves;
	    int       i, numLeaves = autumnGetNumLeaves (s->display);
	    Bool      autumnRotate = autumnGetLeafRotation (s->display);

	    enableTexture (ss->s, &ss->autumnTex[j].tex,
			   COMP_TEXTURE_FILTER_GOOD);

	    for (i = 0; i < numLeaves; i += 2)
	    {
		if (autumnLeaf->tex == &ss->autumnTex[j])
		{
		    glTranslatef (autumnLeaf->x, autumnLeaf->y, autumnLeaf->z);
		
	    	    if (autumnRotate)
			glRotatef (autumnLeaf->ra, 0, 0, 1);
	    	    glCallList (ss->autumnTex[j].dList);
    		    if (autumnRotate)
			glRotatef (-autumnLeaf->ra, 0, 0, 1);
	    	    glTranslatef (-autumnLeaf->x, -autumnLeaf->y, -autumnLeaf->z);
		
		}
		autumnLeaf++;
	    }

	    for (i = 1; i < (numLeaves); i += 2)
	    {
		if (autumnLeaf->tex == &ss->autumnTex[j])
		{
		    glTranslatef (autumnLeaf->x, autumnLeaf->y, autumnLeaf->z);
		
	    	    if (autumnRotate)
			glRotatef (-autumnLeaf->ra, 0, 0, 1);
	    	    glCallList (ss->autumnTex[j].dList);
    		    if (autumnRotate)
			glRotatef (autumnLeaf->ra, 0, 0, 1);
	    	    glTranslatef (-autumnLeaf->x, -autumnLeaf->y, -autumnLeaf->z);
		
		}
		autumnLeaf++;
	    }
	    disableTexture (ss->s, &ss->autumnTex[j].tex);
	}
    }
    else
    {
	Leaf *autumnLeaf = ss->allLeaves;
	int       i, numLeaves = autumnGetNumLeaves (s->display);

	for (i = 0; i < numLeaves; i++)
	{
	    glTranslatef (autumnLeaf->x, autumnLeaf->y, autumnLeaf->z);
	    glRotatef (autumnLeaf->ra, 0, 0, 1);
	    glCallList (ss->displayList);
	    glRotatef (-autumnLeaf->ra, 0, 0, 1);
	    glTranslatef (-autumnLeaf->x, -autumnLeaf->y, -autumnLeaf->z);
	    autumnLeaf++;
	}
    }

    glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
    if (autumnGetUseBlending (s->display))
    {
	glDisable (GL_BLEND);
	glBlendFunc (GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
    }
}

/* ------------------------  FUNCTIONS -------------------- */

static Bool
autumnPaintOutput (CompScreen              *s,
		 const ScreenPaintAttrib *sa,
		 const CompTransform	 *transform,
		 Region                  region,
		 CompOutput              *output, 
		 unsigned int            mask)
{
    Bool status;

    AUTUMN_SCREEN (s);

    if (ss->active && !autumnGetLeavesOverWindows (s->display))
	mask |= PAINT_SCREEN_WITH_TRANSFORMED_WINDOWS_MASK;

    UNWRAP (ss, s, paintOutput);
    status = (*s->paintOutput) (s, sa, transform, region, output, mask);
    WRAP (ss, s, paintOutput, autumnPaintOutput);

    if (ss->active && autumnGetLeavesOverWindows (s->display))
    {
	CompTransform sTransform = *transform;

	transformToScreenSpace (s, output, -DEFAULT_Z_CAMERA, &sTransform);

	glPushMatrix ();
	glLoadMatrixf (sTransform.m);
	beginRendering (ss, s);
	glPopMatrix ();
    }

    return status;
}

static Bool
autumnDrawWindow (CompWindow           *w,
		const CompTransform  *transform,
		const FragmentAttrib *attrib,
		Region               region,
		unsigned int         mask)
{
    Bool status;

    AUTUMN_SCREEN (w->screen);

    /* First draw Window as usual */
    UNWRAP (ss, w->screen, drawWindow);
    status = (*w->screen->drawWindow) (w, transform, attrib, region, mask);
    WRAP (ss, w->screen, drawWindow, autumnDrawWindow);

    /* Check whether this is the Desktop Window */
    if (ss->active && (w->type & CompWindowTypeDesktopMask) && 
	!autumnGetLeavesOverWindows (w->screen->display))
    {
	beginRendering (ss, w->screen);
    }

    return status;
}

static void
initiateLeaf (LeafScreen *ss,
		   Leaf  *sf)
{
    int boxing = autumnGetScreenBoxing (ss->s->display);

    switch (autumnGetWindDirection (ss->s->display))
    {
    case WindDirectionNoWind:
	sf->x  = mmRand (-boxing, ss->s->width + boxing, 1);
	sf->xs = mmRand (-3, 3, 1);
	sf->y  = mmRand (-300, 0, 1);
	sf->ys = mmRand (1, 3, 1);
	break;
    case WindDirectionUp:
	sf->x  = mmRand (-boxing, ss->s->width + boxing, 1);
	sf->xs = mmRand (-3, 3, 1);
	sf->y  = mmRand (ss->s->height, ss->s->height + 300, 1);
	sf->ys = -mmRand (1, 3, 1);
	break;
    case WindDirectionLeft:
	sf->x  = mmRand (ss->s->width, ss->s->width + 300, 1);
	sf->xs = -mmRand (1, 3, 1);
	sf->y  = mmRand (-boxing, ss->s->height + boxing, 1);
	sf->ys = mmRand (-3, 3, 1);
	break;
    case WindDirectionRight:
	sf->x  = mmRand (-300, 0, 1);
	sf->xs = mmRand (1, 3, 1);
	sf->y  = mmRand (-boxing, ss->s->height + boxing, 1);
	sf->ys = mmRand (-3, 3, 1);
	break;
    default:
	break;
    }

    sf->z  = mmRand (-autumnGetScreenDepth (ss->s->display), 0.1, 5000);
    sf->zs = mmRand (-1000, 1000, 500000);
    sf->ra = mmRand (-1000, 1000, 50);
    sf->rs = mmRand (-2100, 2100, 700);
}

static void
setLeafTexture (LeafScreen *ss,
		     Leaf  *sf)
{
    if (ss->autumnTexturesLoaded)
	sf->tex = &ss->autumnTex[rand () % ss->autumnTexturesLoaded];
}

static void
updateLeafTextures (CompScreen *s)
{
    int       i, count = 0;
    float     autumnSize = autumnGetLeafSize(s->display);
    int       numLeaves = autumnGetNumLeaves(s->display);
    Leaf *autumnLeaf;

    AUTUMN_SCREEN (s);
    AUTUMN_DISPLAY (s->display);

    autumnLeaf = ss->allLeaves;

    for (i = 0; i < ss->autumnTexturesLoaded; i++)
    {
	finiTexture (s, &ss->autumnTex[i].tex);
	glDeleteLists (ss->autumnTex[i].dList, 1);
    }

    if (ss->autumnTex)
	free (ss->autumnTex);
    ss->autumnTexturesLoaded = 0;

    ss->autumnTex = calloc (1, sizeof (LeafTexture) * sd->autumnTexNFiles);

    for (i = 0; i < sd->autumnTexNFiles; i++)
    {
	CompMatrix  *mat;
	LeafTexture *sTex;

	ss->autumnTex[count].loaded =
	    readImageToTexture (s, &ss->autumnTex[count].tex,
				sd->autumnTexFiles[i].s,
				&ss->autumnTex[count].width,
				&ss->autumnTex[count].height);
	if (!ss->autumnTex[count].loaded)
	{
	    compLogMessage (s->display, "autumn", CompLogLevelWarn,
			    "Texture not found : %s", sd->autumnTexFiles[i].s);
	    continue;
	}
	compLogMessage (s->display, "autumn", CompLogLevelInfo,
			"Loaded Texture %s", sd->autumnTexFiles[i].s);
	
	mat = &ss->autumnTex[count].tex.matrix;
	sTex = &ss->autumnTex[count];

	sTex->dList = glGenLists (1);
	glNewList (sTex->dList, GL_COMPILE);

	glBegin (GL_QUADS);

	glTexCoord2f (COMP_TEX_COORD_X (mat, 0), COMP_TEX_COORD_Y (mat, 0));
	glVertex2f (0, 0);
	glTexCoord2f (COMP_TEX_COORD_X (mat, 0),
		      COMP_TEX_COORD_Y (mat, sTex->height));
	glVertex2f (0, autumnSize * sTex->height / sTex->width);
	glTexCoord2f (COMP_TEX_COORD_X (mat, sTex->width),
		      COMP_TEX_COORD_Y (mat, sTex->height));
	glVertex2f (autumnSize, autumnSize * sTex->height / sTex->width);
	glTexCoord2f (COMP_TEX_COORD_X (mat, sTex->width),
		      COMP_TEX_COORD_Y (mat, 0));
	glVertex2f (autumnSize, 0);

	glEnd ();
	glEndList ();

	count++;
    }

    ss->autumnTexturesLoaded = count;
    if (count < sd->autumnTexNFiles)
	ss->autumnTex = realloc (ss->autumnTex, sizeof (LeafTexture) * count);

    for (i = 0; i < numLeaves; i++)
	setLeafTexture (ss, autumnLeaf++);
}

static Bool
autumnInitScreen (CompPlugin *p,
		CompScreen *s)
{
    LeafScreen *ss;
    int        i, numLeaves = autumnGetNumLeaves (s->display);
    Leaf  *autumnLeaf;

    AUTUMN_DISPLAY (s->display);

    ss = calloc (1, sizeof(LeafScreen));

    s->privates[sd->screenPrivateIndex].ptr = ss;

    ss->s = s;
    ss->autumnTexturesLoaded = 0;
    ss->autumnTex = NULL;
    ss->active = FALSE;
    ss->displayListNeedsUpdate = FALSE;

    ss->allLeaves = autumnLeaf = malloc (numLeaves * sizeof (Leaf));

    for (i = 0; i < numLeaves; i++)
    {
	initiateLeaf (ss, autumnLeaf);
	setLeafTexture (ss, autumnLeaf);
	autumnLeaf++;
    }

    updateLeafTextures (s);
    setupDisplayList (ss);

    WRAP (ss, s, paintOutput, autumnPaintOutput);
    WRAP (ss, s, drawWindow, autumnDrawWindow);

    ss->timeoutHandle = compAddTimeout (autumnGetAutumnUpdateDelay (s->display),
					stepLeafPositions, s);

    return TRUE;
}

static void
autumnFiniScreen (CompPlugin *p,
		CompScreen *s)
{
    int i;

    AUTUMN_SCREEN (s);

    if (ss->timeoutHandle)
	compRemoveTimeout (ss->timeoutHandle);

    for (i = 0; i < ss->autumnTexturesLoaded; i++)
    {
	finiTexture (s, &ss->autumnTex[i].tex);
	glDeleteLists (ss->autumnTex[i].dList, 1);
    }

    if (ss->autumnTex)
	free (ss->autumnTex);

    if (ss->allLeaves)
	free (ss->allLeaves);

    UNWRAP (ss, s, paintOutput);
    UNWRAP (ss, s, drawWindow);

    free (ss);
}

static void
autumnDisplayOptionChanged (CompDisplay        *d,
			  CompOption         *opt,
			  AutumnDisplayOptions num)
{
    AUTUMN_DISPLAY (d);

    switch (num)
    {
    case AutumnDisplayOptionLeafSize:
	{
	    CompScreen *s;

	    for (s = d->screens; s; s = s->next)
	    {
		AUTUMN_SCREEN (s);
		ss->displayListNeedsUpdate = TRUE;
		updateLeafTextures (s);
	    }
	}
	break;
    case AutumnDisplayOptionAutumnUpdateDelay:
	{
	    CompScreen *s;

	    for (s = d->screens; s; s = s->next)
	    {
		AUTUMN_SCREEN (s);
					
		if (ss->timeoutHandle)
		    compRemoveTimeout (ss->timeoutHandle);
		ss->timeoutHandle =
		    compAddTimeout (autumnGetAutumnUpdateDelay (d),
				    stepLeafPositions, s);
	    }
	}
	break;
    case AutumnDisplayOptionNumLeaves:
	{
	    CompScreen *s;
	    int        i, numLeaves;
	    Leaf  *autumnLeaf;

	    numLeaves = autumnGetNumLeaves (d);
	    for (s = d->screens; s; s = s->next)
    	    {
		AUTUMN_SCREEN (s);
		ss->allLeaves = realloc (ss->allLeaves,
					     numLeaves * sizeof (Leaf));
		autumnLeaf = ss->allLeaves;

		for (i = 0; i < numLeaves; i++)
		{
		    initiateLeaf (ss, autumnLeaf);
    		    setLeafTexture (ss, autumnLeaf);
	    	    autumnLeaf++;
		}
	    }
	}
	break;
    case AutumnDisplayOptionLeafTextures:
	{
	    CompScreen *s;
	    CompOption *texOpt;

	    texOpt = autumnGetLeafTexturesOption (d);

	    sd->autumnTexFiles = texOpt->value.list.value;
    	    sd->autumnTexNFiles = texOpt->value.list.nValue;

	    for (s = d->screens; s; s = s->next)
		updateLeafTextures (s);
	}
	break;
    default:
	break;
    }
}

static Bool
autumnInitDisplay (CompPlugin  *p,
		 CompDisplay *d)
{
    CompOption  *texOpt;
    AutumnDisplay *sd;
    
    sd = malloc (sizeof (AutumnDisplay));

    sd->screenPrivateIndex = allocateScreenPrivateIndex (d);
    if (sd->screenPrivateIndex < 0)
    {
	free (sd);
	return FALSE;
    }
	
    autumnSetToggleInitiate (d, autumnToggle);
    autumnSetNumLeavesNotify (d, autumnDisplayOptionChanged);
    autumnSetLeafSizeNotify (d, autumnDisplayOptionChanged);
    autumnSetAutumnUpdateDelayNotify (d, autumnDisplayOptionChanged);
    autumnSetLeafTexturesNotify (d, autumnDisplayOptionChanged);

    texOpt = autumnGetLeafTexturesOption (d);
    sd->autumnTexFiles = texOpt->value.list.value;
    sd->autumnTexNFiles = texOpt->value.list.nValue;

    d->privates[displayPrivateIndex].ptr = sd;

    return TRUE;
}

static void
autumnFiniDisplay (CompPlugin  *p,
		 CompDisplay *d)
{
    AUTUMN_DISPLAY (d);

    freeScreenPrivateIndex (d, sd->screenPrivateIndex);
    free (sd);
}

static Bool
autumnInit (CompPlugin *p)
{
    displayPrivateIndex = allocateDisplayPrivateIndex ();
    if (displayPrivateIndex < 0)
	return FALSE;

    return TRUE;
}

static void
autumnFini (CompPlugin *p)
{
    freeDisplayPrivateIndex (displayPrivateIndex);
}

static int
autumnGetVersion (CompPlugin *p,
		int        version)
{
    return ABIVERSION;
}

CompPluginVTable autumnVTable = {
    "autumn",
    autumnGetVersion,
    0,
    autumnInit,
    autumnFini,
    autumnInitDisplay,
    autumnFiniDisplay,
    autumnInitScreen,
    autumnFiniScreen,
    0,
    0,
    0,
    0,
    0,
    0
};

CompPluginVTable*
getCompPluginInfo (void)
{
    return &autumnVTable;
}
